/*
    global moment
*/
import eventBus from "../people-view/src/event-bus.js";
import FamilyManager from "./family-manager.js";
import GroupManager from "./group-manager.js";

export default class IMPersonManager
{
    constructor(params)
    {
        this._persons = [];
        this._personsMap = {};
        this._clipboard = {
            type: '',
            persons: []
        }
        this._groups = [];
        this._params = params ? params : {};
        this._categories = [];
        this._loaded = false;
        this._filter = '';
        this._onlyUnconfirmed = false;
        this._imwsInfo = null;
        this._testMode = false;
        this._familyManager = new FamilyManager(this);
        this._groupManager = new GroupManager(this);
        this._personRelationList = {}
        this._focusedPerson = 0;

        if (this._params.unknownPersonsEntity == undefined) {
            this._params.unknownPersonsEntity = true;
        }
    }

    testMode() {
        return this._testMode;
    }

    setTestMode(newMode)
    {
        this._testMode = newMode;
    }

    familyManager() {
        return this._familyManager;
    }

    groupManager() {
        return this._groupManager;
    }

    persons()
    {
        return this._persons;
    }

    /**
     * Returns an array of all ids for the provided persons.
     */
    convertPersonsToIds(persons)
    {
        const res = [];
        for (let p of persons) {
            res.push(p.entity.id);
        }
        return res;
    }

    /**
     * Return all persons not hidden by an active filter.
     */
    getFilteredPersons(scope)
    {
        let res = this._persons.filter(p => {
            return !p.filtered && (scope ? scope.indexOf(p.entity.id) >= 0 : true);
        });
        return res;
    }

    /**
     * Get the label content for the person, or the tag.
     */
    getPersonLabel(id)
    {
        const person = this.getPerson(id);
        const labelProperty = this._imwsInfo.options.faceLabelProperty;
        let res = person.entity.properties[labelProperty];
        if (res && res.length) {
            return res[0];
        }
        return person.entity.tag;
    }

    /**
     * Fetch the person's cover image as a Base64-encoded blob.
     * @param {*} person 
     */
    getPersonCoverImageBase64(id)
    {
        return IMWS.get('v1/person/cover', {
            id: id
        });
    }

    getPersonCoverImageBase64Url(id)
    {
        return `${IMWS.url()}/v1/person/cover?auth_token=${IMWS.authToken()}&id=${id}`;
    }

    /**
     * Sorts persons by the SortBy property, label or fallback property.
     */
    sortPersons(persons) {
        const labelProperty = this._imwsInfo.options.faceLabelProperty;
        persons.sort((personA, personB) => {
            let a = personA.entity.properties['sortExpr'];
            if (!a) {
                a = personA.entity.properties[labelProperty];
                if (!a) {
                    a = personA.entity.tag;
                }
            }
            if (a) {
                // First element or the multi-value elements.
                a = a[0];
            }
            let b = personB.entity.properties['sortExpr'];
            if (!b) {
                b = personB.entity.properties[labelProperty];
                if (!b) {
                    b = personB.entity.tag;
                }
            }
            if (b) {
                b = b[0];
            }

            if (a && b) {
                return a.localeCompare(b);
            }
            else {
                return personA.entity.id < personB.entity.id;
            }
        });
    }

    groups()
    {
        return this._groups;
    }


    loaded()
    {
        return this._loaded;
    }

    imwsInfo() {
        return this._imwsInfo;
    }

    clipboardCopy(type, persons)
    {
        this._clipboard.type = type;
        this._clipboard.persons = persons ? persons : [];
    }

    clipboardData()
    {
        return this._clipboard;
    }

    clipboardEmpty()
    {
        return this._clipboard.persons.length == 0;
    }

    /**
     * Get the person with the specified id.
     * @param {*} id 
     */
    getPerson(id)
    {
        return this._personsMap[id];
    }

    /**
     * Return the 0-based index of the specified person or -1, if the 
     * person does not exist.
     * @param {*} id 
     * @returns {Number}
     */
    getPersonIndex(id)
    {
        for (let i = 0; i < this._persons.length; ++i) {
            if (this._persons[i].entity.id == id) return i;
        }
        return -1;
    }

    getGroup(name)
    {
        for (let i = 0; i < this._groups.length; i++) {
            if (this._groups[i].name == name) {
                return this._groups[i];
            }
        }
        return undefined;
    }

    // Calculate the age of the specified person in years.
    // Returns undefined when no birth date exists.
    // If the entiry has a DOD, calculates the age between
    // DOB and DOD.
    calculateAge(entity)
    {
        const b = this.getProperty(entity,'dob');
        const d = this.getProperty(entity,'dod');
        if (b && d) {
            return this.calculateAgeBetween(moment(this.getProperty(entity,'dob')),d ? moment(d) : undefined,'years');
        } else if (b) {
            return this.calculateAgeBetween(moment(this.getProperty(entity,'dob')), undefined,'years');
        } else {
            return undefined;
        }
    }

    // Calculate the age between dob and dod.
    // If dod is undefined, calculates the age based on today.
    // Returns undefined if dob and dod are undefined.
    calculateAgeBetween(dob,dod,unit)
    {
        let res;
        if (dob) {
            if (dod) {
                res = dod.diff(dob,unit);
            }
            else {
                res = moment().diff(dob,unit);
            }
        }
        return res;
    }

    /**
     * Group persons by the specified criteria and return an 
     * object with a key for each criteria and the persons matching that key.
     * @param {string} propertyName
     * @returns 
     */
    groupPersons(propertyName) {
        const groups = {};

        for (let p of this._persons) {
            const g = p.entity.properties[propertyName];
            if (g) {
                if (groups[g] === undefined) {
                    groups[g] = [];
                }
                groups[g].push(p);
            }
        }        

        return groups;
    }

    /**
     * Return the first (!) property or undefined if the property
     * does not exist.
     * @param {*} e The entity object
     * @param {*} name The name of the property to retrieve.
     * @param {*} index The 0-based index of the property to retrieve. If this is -1, all property values are concatenated by sep and returned.
     */
    getProperty(e,name,index = 0, sep = '')
    {
        if (e.properties[name] && e.properties[name].length) {
            if (index >= 0) {
                if (index < e.properties[name].length) {
                    return e.properties[name][index];
                }
                else {
                    return undefined;
                }
            }
            else {
                // Concatenate
                let res = '';
                for (let p of e.properties[name]) {
                    if (res.length) res += sep;
                    res += p;
                }
                return res;
            }
        }
        else {
            return undefined;
        }
    }

    /**
     * Return the property array or undefined if the property
     * does not exist.
     * @param {*} e The entity object
     * @param {*} name The name of the property to retrieve.
     */
    getProperties(e,name)
    {
        if (e.properties[name]) {
            return e.properties[name];
        }
        else {
            return undefined;
        }
    }

    /**
     * Return the category path for the specified category.
     */
    getCategoryPathForCategoryId(id)
    {
        for (const c of  this._categories) {
            if (c.id == id) {
                return c.path;
            }
        }
        return undefined;
    }

    /**
     * Switch to the category view and select the category with the specified id.
     */
    gotoCategory(id)
    {
        IMatch.gotoCategory(id);
    }

    /**
     * Switch to the category view and select the @Keywords category
     * matching the keyword.
     */
    async gotoKeyword(keyword)
    {
        try {
            // person keywords contain group and exclude levels.
            // The corresponding @Keywords category does not include group levels.
            // This means we need to map the keyword first to get rid of the group levels.
            let result = await IMWS.get('v1/thesaurus/explodekeywords', {
                keywords: keyword,
                includesgrouplevels: true,
                includeexcludes: true
            });

            // The result contains the hierarchical keywords as they would be assigned to a file,
            // without the group levels.
            const r = result[keyword];

            // Use this to lookup the corresponding @Keywords category.
            if (r && r.keywords && r.keywords.length) {
                let categories = await IMWS.get('v1/categories',{
                    path: `@Keywords|${r.keywords[0]}`,
                    fields: 'id'
                });

                // And switch to it.
                if (categories.categories.length) {
                    IMatch.gotoCategory(categories.categories[0].id);
                }
            }
        } catch (error) {
            console.error(error);
        }
    }

    /**
     * Set the focus to the specified person.
     * id = 0 means the "Unknown Person".
     * If id is -1, the focus is removed.
     * @param {*} id 
     */
    setFocus(id)
    {
        this._focusedPerson = id !== undefined ? id : -1;
    }

    /**
     * Return the id of the currently focused person.
     * If 0 is returned, the "Unknown Person" is focused.
     * Returns -1 if there is no focus.
     */
    getFocus()
    {
        return this._focusedPerson;
    }

    /**
     * Return true if the specified person is selected.
     * @param {*} id Person id
     * @returns {Boolean}
     */
    isSelected(id)
    {
        for (let i = 0; i < this._persons.length; ++i) {
            if (this._persons[i].entity.id == id) {
                return this._persons[i].selected;
            }
        }
        return false;
    }


    /**
     * Change the selection state of an event
     * @param {Number} id Person id
     * @param {Boolean} selected New selection state
     */
    select(id,selected)
    {
        for (let i = 0; i < this._persons.length; ++i) {
            if (this._persons[i].entity.id == id) {
                if (this._persons[i].selected !== selected) {
                    this._persons[i].selected = selected;
                }
                break;
            }
        }
    }

    /**
        Select all persons.
    */
    selectAll()
    {
        for (let p of this._persons) {
            if (!p.selected) {
                p.selected = true;
            }
        }
    }

    /**
     * Deselect all persons.
     */
    clearSelection()
    {
        for (let p of this._persons) {
            // Avoid too many unneeded updates => watches / reactivity.
            if (p.selected) {
                p.selected = false;
            }
        }
    }

    /**
     * Get all selected persons.
     * @returns array of person ids
     */
    getSelection()
    {
        let res = [];
        for (let p of this._persons) {
            if (p.selected) res.push(p.entity.id);
        }
        return res;
    }

    /**
     * 
     * @param {string} filter 
     */
    setFilter(pattern)
    {
        if (this._filter != pattern) {
            this._filter = pattern;
            this._search(this._filter,true);
        }
    }

    /**
     * Show only unfonfirmed persons?
     * @param {string} enable 
     */
    setShowOnlyUnconfirmed(enable)
    {
        if (this._onlyUnconfirmed != enable) {
            this._onlyUnconfirmed = enable;
            this._search(this._filter,true);
        }
    }

    /**
     * Returns true if all persons without unconfirmed fasces are filtered.
     */
    showOnlyUnconfirmed()
    {
        return this._onlyUnconfirmed;
    }

    /**
     * Load all people groups from the database.
     */
    loadGroups()
    {
        this._groups = [];
        return new Promise((resolve,reject) => {
            IMWS.get('v1/entities',{
                schema: 'people_group'
            }).then(response => {
                for (let e of response.entities) {
                    this._groups.push({
                        entity: e,
                        name: this.getProperty(e,'name') ? this.getProperty(e,'name') : e.tag,
                        // We keep the original name as loaded.
                        // This allows us to tell if the group was renamed.
                        originalName: this.getProperty(e,'name') ? this.getProperty(e,'name') : e.tag,
                        color: this.getProperty(e,'color')
                    });
                }
                this._groups.sort((a,b) => {
                    return a.name.localeCompare(b.name);
                });
                resolve();
            },error => {
                reject(error);
            });
        });
    }

    deleteGroup(nameOfGroupToDelete)
    {
        for (let i = 0; i < this._groups.length; i++) {
            if (this._groups[i].name === nameOfGroupToDelete) {
                this._groups.splice(i,1);
                this.saveGroups();
                break;
            }
        }
    }

    /**
     * Add the persons specified in PersonIds to the group groupMame.
     * If removeFromGroupName is specified, the persons are removed
     * from that group.
     * @param {*} groupName The group to add the persons to.
     * @param {*} personIds Array with person ids.
     * @param {*} removeFromGroupName The optional group to remove the persons from.
     */
    async addPersonsToGroup(groupName, personIds, removeFromGroupName) {
        if (personIds.length) {
            try {
                const entitiesToUpdate = [];
                // Fetch the entities for all persons dropped.
                const response = await IMWS.get('v1/entities', {
                    schema: 'people',
                    id: personIds.join(',')
                });
                response.entities.forEach((person) => {
                    const cachedPerson = this.getPerson(person.id);
                    let groups = person.properties['group'];
                    if (!groups) {
                        if (cachedPerson) {
                            cachedPerson.entity.properties['group'] = [groupName];
                        }
                        person.schema = 'people';
                        person.properties['group'] = [groupName];
                        entitiesToUpdate.push(person);
                    }
                    else {
                        // Add the group to the person if needed.
                        let updated = false;

                        let index = groups.indexOf(groupName);
                        if (index < 0) {
                            groups.push(groupName);
                            updated = true;
                        }

                        if (removeFromGroupName) {
                            let index = groups.indexOf(removeFromGroupName);
                            if (index >= 0) {
                                groups.splice(index, 1);
                                updated = true;
                            }
                        }

                        if (updated) {
                            if (cachedPerson) {
                                cachedPerson.entity.properties['group'] = groups;
                            }
                            person.schema = 'people';
                            entitiesToUpdate.push(person);
                        }
                    }
                });

                if (entitiesToUpdate.length) {
                    await IMWS.post('v1/entities', {
                        data: JSON.stringify(entitiesToUpdate)
                    });
                }
            }
            catch (e) {
                console.error(e);
                this.load();
            }
        }
    }

    setGroups(newGroups)
    {
        this._groups = newGroups;
    }

    saveGroups()
    {
        let data = [];

        // Do we have renamed groups?
        // If yes, we need to update all persons with the original group
        // to use the new group name
        let renamed = [];
        for (let g of this._groups) {
            if (g.originalName && g.originalName != g.name) {
                renamed.push({
                    o: g.originalName,
                    n: g.name
                });
            }
        }
        
        for (let g of this._groups) {
            data.push({
                schema: 'people_group',
                tag: g.tag,
                properties: {
                    color: g.color,
                    name: g.name
                }
            });
        }

        // Delete all groups
        IMWS.delete('v1/entities',{
            schema: 'people_group',
            id: 'all'
        }).then(() => {
            // Set the new groups
            IMWS.post('v1/entities',{
                data: JSON.stringify(data)
            }).then(() => {
                // Load the new groups, make sure originalName is correct.
                this.loadGroups();
                // Update entities for renamed groups

                for (let r of renamed) {
                    return IMWS.post('v1/entities/property',{
                        schema: 'people',
                        property: 'group',
                        values: JSON.stringify([
                            r.n
                        ]),
                        filters: JSON.stringify([
                            {
                                type: 'property',
                                op: 'is',
                                name: 'group',
                                value: r.o
                            }
                        ])
                    });
                };
            });
        });
    }

    /**
     * Clone the specified person.
     * @param {*} id 
     * @returns {Object} Response object
     */
    clonePerson(id)
    {
        return new Promise((resolve,reject) => {
            IMWS.get('v1/entities',{
                schema: 'people',
                id: id,
                noblob: false
            }).then(response => {
                if (response.entities.length) {
                    let e = response.entities[0];

                    // We just create a new unique tag and store the entity back.
                    IMWS.get('v1/entity/tag/getunique',{
                        schema: 'people',
                        tag: e.tag
                    }).then(tagData => {
                        e.id = 0;
                        e.tag = tagData.tag;
                        e.schema = 'people';
                        IMWS.post('v1/entities',{
                            data: JSON.stringify([e])
                        }).then(result => {
                            resolve(result);
                        },error => {
                            reject(error);
                        });
                    },error => {
                        reject(error);
                    });
                }
            },error => {
                reject(error);
            });
        });
    }

    /**
     * Reload all people, groups and families from the database.
     */
    async load()
    {
        this._imwsInfo = await IMWS.get('info');

        this._categories = [];
        await this.loadGroups();

        // We restore the selection (if any) after the load.
        let selection = [];
        for (let p of this._persons) {
            if (p.selected) {
                selection.push(p.entity.id);
            }
        }

        // We restore the filter (if any) after the load.
        let filtered = [];
        for (let p of this._persons) {
            if (p.filtered) {
                filtered.push(p.entity.id);
            }
        }
        
        this._loaded = false;
        this._persons = [];
        this._personsMap = {};

        const response = await IMWS.get('v1/entities',{
            schema: 'people',
            sortby: 'sortExpr:asc',
            // We don't load blob data (including the cover file) here.
            // We load this data (and file counts) only on-demand.
            noblob: true
        });

        // Get all supported 

        this._personRelationList = await IMWS.get('v1/person/relationlist');

        this._loaded = true;
        for (let e of response.entities) {
            const person = {
                entity: e,
                selected : selection.indexOf(e.id) >= 0,
                filtered : filtered.indexOf(e.id) >= 0,
                fileCount : undefined,
                age: 0
            }
            Vue.set(person,'coverKey',0);
            this._persons.push(person);
            this._personsMap[person.entity.id] = person;
        }

        // Append the special "Unknown Person" entity.
        if (this._params.unknownPersonsEntity) {
            const person = {
                entity: {
                    id: 0,
                    tag: IMatchTranslate.getStr('T_TAG_UNKNOWN_PERSONS'),
                    properties: {}
                },
                selected : selection.indexOf(0) >= 0,
                filtered: filtered.indexOf(0) >= 0,
                fileCount : undefined,
                age: 0
            }
            Vue.set(person,'coverKey',0);
            this._persons.push(person);
            this._personsMap[0] = person;
        }

        let now = moment();
        for (let p of this._persons) {
            // If we have a DOB, calculate the age.
            let d = this.getProperty(p.entity,'dob');
            if (d) {
                p.age = this.calculateAge(p.entity);
            }
        }

        // Get all categories for the loaded persons.
        const categoriesToLoad = [];
        for (let p of this._persons) {
            const cats = p.entity.properties['categories'];
            if (cats && cats.length) {
                cats[0].forEach(c => {
                    if (categoriesToLoad.indexOf(c) < 0) {
                        categoriesToLoad.push(c);
                    }
                });
            }
        }

        if (categoriesToLoad.length) {
            const response = await IMWS.get('v1/categories',{
                id: categoriesToLoad.join(','),
                fields: 'id,path'
            });
            this._categories = response.categories;
        }

        await Promise.all([this._familyManager.load(), this._groupManager.load()]);
        eventBus.$emit('persons-loaded');
    }

    /**
     * Load one person from the database.
     * Update the existing person or add a new person.
     * @param {*} id 
     */
    loadPerson(id, loadBlobs = false)
    {
        return new Promise((resolve,reject) => {
            IMWS.get('v1/entities',{
                schema: 'people',
                id: id,
                // If this is false, IMWS returns an array with one empty element [''] in 
                // entity.properties.coverFile when the person has a cover file.
                noblob: !loadBlobs
            }).then(response => {
                let person;
                if (response.entities.length) {
                    let e = response.entities[0];

                    if (this._personsMap[e.id]) {
                        this._personsMap[e.id].entity = e;
                        person = this._personsMap[e.id];
                    }
                    if (!person) {
                        person = {
                            entity: e,
                            fileCount: undefined,
                            selected: false,
                            filtered: false,
                            age: 0,
                        }
                        this._persons.push(person);
                        this._personsMap[e.id] = person;
                    }
                    person.age = this.calculateAge(person.entity);
                }

                // We use this to trigger cover file load / reload when needed.
                Vue.set(person,'coverKey',0);
                
                resolve(person);
            }, error => {
                reject(error);
            });
        });
    }

    /**
     * Calculate/update the file count for the specified person.
     * Use id = 0 to update the unknown persons.
     * @param {*} id 
     */
    calculateFiles(id)
    {
        return new Promise((resolve,reject) => {

            if (id == 0) {
                if (this._params.unknownPersonsEntity) {
                    IMWS.get('v1/person/files/unassigned',{
                        fields: 'id'
                    }).then(response => {
                        let person = this._personsMap[id];
                        if (person) {
                            if ((person.fileCount === undefined) || (person.fileCount !== response.files.length)) {
                                person.fileCount = response.files.length;
                            }
                        }
                        resolve();
                    }, error => {
                        reject(error);
                    });
                }
            }
            else {
                IMWS.get('v1/person/files/count',{
                    id: id
                }).then(response => {
                    let person = this._personsMap[id];
                    if (person) {
                        if ((person.fileCount === undefined) || (person.fileCount !== response)) {
                            person.fileCount = response;
                        }
                    }
                    resolve();
                }, error => {
                    reject(error);
                });
            }
        });
    }

    /**
     * Calculate the files for the specified person.
     * This also works if the person has not been loaded before (different than calculateFiles()
     * @param {Number} id Id of the person
     * @returns {Promise} If resolved, the number of files for the person. 
     */
    getFileCount(id)
    {
        return new Promise((resolve,reject) => {
            IMWS.get('v1/person/files/count',{
                id: id
            }).then(response => {
                resolve(response);
            }, error => {
                reject(error);
            });
        });
    }

    loadFaceData(id)
    {
        return new Promise(async (resolve,reject) => {
            if (id == 0) {
                // The unknown person
                resolve();
            } else {
                // Now load the face info
                try {
                    const response = await IMWS.get('v1/person/facecounts',{
                        id: id,            
                    });
                    let person = this._personsMap[id];
                    if (person) {
                        let modified = false;
                        const ucc = response.faceCount - response.confirmed;
                        if (person.unconfirmedCount !== ucc) {
                            person.unconfirmedCount = ucc;
                        }
                        if (person.rejectedCount !== response.rejected) {
                            person.rejectedCount = response.rejected;
                        }
                        if (person.trainedCount !== response.train) {
                            person.trainedCount = response.train;
                        }
                    }
                    resolve();
                }
                catch (error) {
                    reject(error);
                }
            }
        });
    }

    getCoverFileUrl(id)
    {
        return `${IMWS.url()}/v1/person/cover?auth_token=${IMWS.authToken()}&id=${id}`;
    }

    /**
     * Calculate the age of the person. If no DOB is specified, returns undefined.
     * @param {*} id The person id
     */
    getAge(id)
    {
        let p = this.getPerson(id);
        if (p) {
            let now = moment();
            let d = this.getProperty(p.entity,'dob');
            if (d) {
                return now.diff(moment(d),'years');
            }
        }
        return undefined;
    }

    /**
     * Search all persons for the sepecified pattern.
     * @param {*} pattern 
     * @returns {Array} Array of matching persons.
     */
    search(pattern)
    {
        return this._search(pattern,false);
    }

    _search(pattern,setFiltered = false)
    {
        let res = [];

        if (!pattern || !pattern.length) {
            if (setFiltered) {
                for (let p of this._persons) {
                    if (this._onlyUnconfirmed) {
                        p.filtered = p.unconfirmedCount == 0;
                    }
                    else {
                        p.filtered = false;
                    }
                }
            }
            return res;
        }

        /*
            We support the following prefixes:
            t:<pattern>     : Search tags only
            p:<pattern>     : Search person properties only
            l:<pattern>     : Search person label only
            g:<pattern>     : Search groups only
            f:<pattern>     : Search family names only
            <pattern>       : Search anywhere
        */

        let prefix = null;
        const prefixIndex = pattern.indexOf(':');
        if (prefixIndex > 0) {
            prefix = pattern.substring(0,prefixIndex).toLowerCase();
            pattern = pattern.substring(prefixIndex + 1);
        }

        if (pattern.length == 0) {
            return res;
        }

        // The properties to search.
        // If a property contains :s, we match from the start.
        // This handles cases like when searching for 'male' to avoid finding female.
        const PROPERTIES = ['fullName','shortName','secondaryName','description','gender:s'];

        let text = IMatch.normalizeString(pattern);
        let startMatch;
        let match;
        let added = false;
        for (let p of this._persons) {
            p.filtered = true;
            added = false;

            // Tag

            if (!prefix || (prefix === 't')) {
                if (IMatch.normalizeString(p.entity.tag).indexOf(text) >= 0) {
                    if (!added) {
                        res.push(p);
                        added = true;
                    }
                    if (setFiltered) {
                        p.filtered = false;
                    }
                }
            }

            // Person label

            if (p.filtered && (!prefix || (prefix == 'l'))) {
                if (IMatch.normalizeString(this.getPersonLabel(p.entity.id)).indexOf(text) >= 0) {
                    if (!added) {
                        res.push(p);
                        added = true;
                    }
                    if (setFiltered) {
                        p.filtered = false;
                    }
                }
            }

            // Persons

            if (p.filtered && (!prefix || (prefix == 'p'))) {
                for (let name of PROPERTIES) {
                    if (name.indexOf(':s') >= 0) {
                        startMatch = true;
                        name = name.replace(':s','');
                    }
                    else {
                        startMatch = false;
                    }
                    let property = this.getProperty(p.entity,name,-1,'\n');
                    if (property) {
                        if (!startMatch) {
                            match = IMatch.normalizeString(property).indexOf(text) >= 0;
                        }
                        else {
                            match = IMatch.normalizeString(property).indexOf(text) == 0;
                        }
                        if (match) {
                            if (!added) {
                                res.push(p);
                                added = true;
                            }
                            if (setFiltered) {
                                p.filtered = false;
                            }
                            break;
                        }
                    }
                }
            }
                
            // Groups

            if (p.filtered && (!prefix || (prefix == 'g'))) {
                let groupIds = this.groupManager().getGroupsForPerson(p.entity.id);
                if (groupIds.length) {
                    for (let id of groupIds) {
                        const g = this.groupManager().getGroup(id);
                        if (g) {
                            let property = this.getProperty(g.entity,'name',0);
                            if (property) {
                                match = IMatch.normalizeString(property).indexOf(text) >= 0;
                                if (match) {
                                    if (!added) {
                                        res.push(p);
                                        added = true;
                                    }
                                    if (setFiltered) {
                                        p.filtered = false;
                                    }
                                    break;
                                }
                            }
                        }
                    }
                }
            }

            // Families

            if (p.filtered && (!prefix || (prefix == 'f'))) {
                let families = this.familyManager().getFamiliesForPerson(p.entity.id);
                if (families.length) {
                    for (let family of families) {
                        let property = this.getProperty(family.entity,'name',0);
                        if (property) {
                            match = IMatch.normalizeString(property).indexOf(text) >= 0;
                            if (match) {
                                if (!added) {
                                    res.push(p);
                                    added = true;
                                }
                                if (setFiltered) {
                                    p.filtered = false;
                                }
                                break;
                            }
                        }
                    }
                }
            }

            if (!p.filtered) {
                if (this._onlyUnconfirmed && (p.unconfirmedCount == 0)) {
                    p.filtered = true;
                }
            }
        }
        return res;
    }

    /**
     * Retrieve all outgoing edges for the person for the 'person' schema.
     * Returns an array of persons this person is linked to.
     */
    getPersonRelations(id)
    {
        return IMWS.get('v1/person/relations', {
            id: id
        });
    }

    /**
     * Sorts the relations based on their sort key and localized name.
     * @param {*} relations 
     */
    sortPersonRelations(relations)
    {
        relations.sort((a,b) => {
            if (a.auto == b.auto) {
                if (a.sortKey < b.sortKey) {
                    return -1;
                }
                else if (a.sortKey > b.sortKey) {
                    return +1;
                }
                else {
                    return a.localizedName.localeCompare(b.localizedName);
                }
            }
            // Sort auto entries to the bottom.
            else if (a.auto && !b.auto) {
                return +1;
            }
            else {
                return -1;
            }
        });
    }

    /**
     * Update the relations for the specified person by 
     * updating all to's.
     * @param {*} relations Set of relations
     * @returns 
     */
    async setPersonRelations(relations)
    {
        for (const r of relations) {
            if (r.modified !== true) continue;

            if (r.deleted === true) {
                if (r.id !== 0) {
                    await IMWS.delete('v1/person/relations', {
                        id: r.id
                    });
                }
            }
            else if (r.confirmed === true) {
                // Confirm an auto edge.
                await IMWS.post('v1/person/relations/add', {
                    from: r.from,
                    to: r.to,
                    type: r.type
                });
            }
            else if (r.created === true) {
                // New edge 
                await IMWS.post('v1/person/relations/add', {
                    from: r.from,
                    to: r.to,
                    type: r.type
                });
            }
            else {
                await IMWS.post('v1/person/relations/update', {
                    id: r.id,
                    to: r.to,
                    type: r.type
                });
            }
        }
    }

    /**
     * Return a list of all standard person relation types ("father", "sister", ...)
     * supported by the Family Manager. The key is the internal id to use,
     * the localizedName is the name to show to the user.
     * Person edges use the relation key in their "type" property.
     * @returns 
     */
    getPersonRelationList()
    {
        return this._personRelationList;
    }
}

